(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import AdditionalTactics.
Require Import Misc.
Require Import Fomega.
Require Import Derived.
Require Import Modules_syntax.


(*************************************************************************)
(* Semantic signatures *)
(*************************************************************************)

Inductive sig : Set :=
  | valS : typ -> sig (* [t] *)
  | typS : typ -> knd -> sig (* [=t:k] *)
  | sigS : qsig -> sig (* [=xi:O] *)
  | strS : str -> sig (* {ls:Ss} *)
  | funS : list knd -> sig -> qsig ->  sig (* !as.S->X *)

with str : Set :=
  | empR : str
  | labR : lab -> sig -> str -> str

with qsig : Set :=
  | exQ : list knd -> sig -> qsig (* ?as.S *)
.

Scheme IIS := Induction for sig Sort Prop
with IIR := Induction for str Sort Prop
with IIQ := Induction for qsig Sort Prop.
Combined Scheme IISRQ from IIS, IIR, IIQ.


(*************************************************************************)
(* Structures *)
(*************************************************************************)

Fixpoint sizeR r {struct r} : nat :=
  match r with
  | empR => 0
  | labR l s r' => 1 + sizeR r'
  end.

Fixpoint labsR r {struct r} : list lab :=
  match r with
  | empR => nil
  | labR l s r' => l :: labsR r'
  end.

Fixpoint find (l : lab) (r : str) {struct r} : option sig :=
  match r with
  | empR => None
  | labR l' s r' => if l === l' then Some s else find l r'
  end.

Definition field l s r := find l r = Some s.

Fixpoint lfoldR A f (a : A) r {struct r} : A :=
  match r with
  | empR => a
  | labR l s r' => f l (lfoldR A f a r')
  end.
Implicit Arguments lfoldR [A].

Fixpoint seqR r1 r2 {struct r1} : str :=
  match r1 with
  | empR => r2
  | labR l s r1' => labR l s (seqR r1' r2)
  end.

Fixpoint diffR r1 r2 {struct r1} : str :=
  match r1 with
  | empR => empR
  | labR l s r1' =>
    match find l r2 with
    | None => labR l s (diffR r1' r2)
    | Some _ => diffR r1' r2
    end
  end.

Fixpoint interR r1 r2 {struct r1} : list (lab * sig * sig) :=
  match r1 with
  | empR => nil
  | labR l s1 r1' =>
    match find l r2 with
    | None => interR r1' r2
    | Some s2 => (l, s1, s2) :: interR r1' r2
    end
  end.


Lemma lfoldR_length : forall A (f : lab -> A) r,
  length (lfoldR (fun l xs => f l :: xs) nil r) = sizeR r.
Proof.
  intros A f.
  induction r; simpl sizeR; simpl lfoldR; simpl length.
    reflexivity.
    f_equal. apply IHr.
Qed.


(*************************************************************************)
(* Module environments *)
(*************************************************************************)

Notation menv := (list (Id * sig * var)).
Notation "X ~' s @ x" := ((X, s, x)::nil : menv) (at level 31).
Notation "G1 &' G2" := (G2 ++ G1 : menv) (at level 40).

Fixpoint get' (X : Id) (G : menv) {struct G} : option (sig * var) :=
  match G with
  | nil => None
  | (X',s,x) :: G' => if X === X' then Some (s, x) else get' X G'
  end.

Definition binds' X s x G := get' X G = Some (s, x).

(* add a structure to the context *)
Fixpoint bind_str r xs {struct r} : menv := 
  match r, xs with
   | empR, nil => nil
   | labR X s r', x::xs' => (X, s, x) :: bind_str r' xs'
   | _, _ => nil
  end.
Notation "r @@ xs" := (rev (bind_str r xs)) (at level 31).


(*************************************************************************)
(* Erasure into F types *)
(*************************************************************************)

Fixpoint transS  (s : sig) {struct s} : typ  :=
  match s with
    | valS t => t
    | typS t k => typT t k
    | sigS q => typT (transQ q) omK
    | strS r => strT (transR r)
    | funS ks s' q => allsT ks (arrT (transS s') (transQ q))
  end

with transR r {struct r} : row typ :=
   match r with
   | empR => nilR
   | labR l s r' => consR l (transS s) (transR r')
   end

with transQ q {struct q} : typ := 
   match q with
   | exQ ks s => exsT ks (transS s) 
   end
.

Fixpoint transG (G : menv) {struct G} : eenv :=
  match G with
  | nil => empty
  | (X, s, x)::G' => transG G' & x~(transS s)
  end.


Lemma transR_seq : forall r1 r2,
  transR (seqR r1 r2) = catR (transR r1) (transR r2).
Proof.
 induction r1; intros r2.
  reflexivity.
  simpl.  rewrite <- IHr1. reflexivity.
Qed.

Lemma labsR_transR : forall r,
  labsR r = Derived_strT.labsR (transR r).
Proof.
  induction r; simpl transR; simpl labsR; simpl Derived_strT.labsR.
    reflexivity.
    f_equal. apply IHr.
Qed.

Lemma getR_transR : forall l r,
  getR l (transR r) =
    match find l r with
    | None => None
    | Some s => Some (transS s)
    end.
Proof.
  intro l.
  induction r; simpl.
    reflexivity.
    destruct (l === l0). reflexivity. apply IHr.
Qed.

Lemma transR_diff : forall r1 r2,
  transR (diffR r1 r2) = Derived_strT.diffR (transR r1) (transR r2).
Proof.
 induction r1; intros r2.
  reflexivity.
  simpl. rewrite getR_transR. destruct (find l r2); simpl.
    apply IHr1.
    rewrite IHr1. reflexivity.
Qed.

Lemma transG_concat : forall G1 G2,
  transG (G1 &' G2) = transG G1 & transG G2.
Proof.
  intro G1.
  induction G2; simpl transG; simpl.
    reflexivity.
    destruct a as [[X s] x]. rewrite IHG2. reflexivity.
Qed.

Lemma transG_str : forall r xs,
  transG (r@@xs) = xs~~fieldsR (transR r).
Proof.
  induction r; intros xs; simpl transR; simpl fieldsR; simpl bind_str; destruct xs.
    simpl. reflexivity.
    simpl. reflexivity.
    simpl. reflexivity.
    simpl. rewrite transG_concat. rewrite IHr. reflexivity.
Qed.

Lemma transR_length : forall r,
  lengthR (transR r) = sizeR r.
Proof.
  induction r; simpl transR; simpl lengthR; simpl sizeR.
    reflexivity.
    f_equal. apply IHr.
Qed.


(*************************************************************************)
(* Substitution *)
(*************************************************************************)

Fixpoint substAS (a : occ) (t : typ) (s : sig) {struct s} : sig :=
  match s with
  | valS t' => valS (substAT a t t')
  | typS t' k => typS (substAT a t t') k
  | sigS q => sigS (substAQ a t q)
  | strS r => strS (substAR a t r)
  | funS ks s' q => funS ks (substAS (shifts (length ks) a) t s') (substAQ (shifts (length ks) a) t q)
  end

with substAR (a : occ) (t : typ) r {struct r} : str :=
   match r with
   | empR => empR
   | labR l s r' => labR l (substAS a t s) (substAR a t r')
   end

with substAQ (a : occ) (t : typ) q {struct q} : qsig := 
   match q with
   | exQ ks s => exQ ks (substAS (shifts (length ks) a) t s)
   end
.


Fixpoint substsAS (as' : list occ) (ts : list typ) s {struct as'} : sig :=
  match as' with
  | nil => s
  | a::as'' =>
    match ts with
    | nil => s
    | t::ts' => substAS a t (substsAS as'' ts' s)
    end
  end.

Fixpoint substsAR (as' : list occ) (ts : list typ) r {struct as'} : str :=
  match as' with
  | nil => r
  | a::as'' =>
    match ts with
    | nil => r
    | t::ts' => substAR a t (substsAR as'' ts' r)
    end
  end.

Fixpoint substsAQ (as' : list occ) (ts : list typ) q {struct as'} : qsig :=
  match as' with
  | nil => q
  | a::as'' =>
    match ts with
    | nil => q
    | t::ts' => substAQ a t (substsAQ as'' ts' q)
    end
  end.


Lemma substASRQ_fresh_irrel : forall t1 t2,
  (forall s a1 a2,
   a1 <> a2 ->
   substAS a1 t1 s = substAS a2 t2 (substAS a1 t1 s) ->
   substAS a2 t2 s = s) /\
  (forall r a1 a2,
   a1 <> a2 ->
   substAR a1 t1 r = substAR a2 t2 (substAR a1 t1 r) ->
   substAR a2 t2 r = r) /\
  (forall q a1 a2,
   a1 <> a2 ->
   substAQ a1 t1 q = substAQ a2 t2 (substAQ a1 t1 q) ->
   substAQ a2 t2 q = q).
Proof.
  intros t1 t2.
  unfold substAS, substAR, substAQ.
  apply IISRQ.
    Case "valS".
      intros t a1 a2 Neq Hs.
      inversion Hs. f_equal.
      eapply substAT_substAT_irrel. apply Neq. apply H0.
    Case "typS".
      intros t k a1 a2 Neq Hs.
      inversion Hs. f_equal.
      eapply substAT_substAT_irrel. apply Neq. apply H0.
    Case "sigS".
      intros q IHq a1 a2 Neq Hs.
      inversion Hs. f_equal.
      eapply IHq. apply Neq. apply H0.
    Case "strS".
      intros r IHr a1 a2 Neq Hs.
      inversion Hs. f_equal.
      eapply IHr. apply Neq. apply H0.
    Case "funS".
      intros ks s IHs q IHq a1 a2 Neq Hs.
      inversion Hs. f_equal.
      eapply IHs with (a1 := shifts (length ks) a1).
        intro Eq. apply (Neq (shifts_injectivity (length ks) a1 a2 Eq)).
        apply H0.
      eapply IHq with (a1 := shifts (length ks) a1).
        intro Eq. apply (Neq (shifts_injectivity (length ks) a1 a2 Eq)).
        apply H1.
    Case "empR".
      intros a1 a2 Neq Hr.
      reflexivity.
    Case "labR".
      intros l s IHs r IHr a1 a2 Neq Hr.
      inversion Hr. f_equal.
      eapply IHs. apply Neq. apply H0.
      eapply IHr. apply Neq. apply H1.
    Case "exQ".
      intros ks s IHs a1 a2 Neq Hq.
      inversion Hq. f_equal.
      eapply IHs with (a1 := shifts (length ks) a1).
        intro Eq. apply (Neq (shifts_injectivity (length ks) a1 a2 Eq)).
        apply H0.
Qed.

Corollary substAS_fresh_irrel : forall a1 t1 a2 t2 s,
  a1 <> a2 ->
  substAS a1 t1 s = substAS a2 t2 (substAS a1 t1 s) ->
  substAS a2 t2 s = s.
Proof.
  intros a1 t1 a2 t2 s Neq Hs.
  generalize s a1 a2 Neq Hs.
  eapply proj1. apply substASRQ_fresh_irrel.
Qed.

Corollary substAR_fresh_irrel : forall a1 t1 a2 t2 r,
  a1 <> a2 ->
  substAR a1 t1 r = substAR a2 t2 (substAR a1 t1 r) ->
  substAR a2 t2 r = r.
Proof.
  intros a1 t1 a2 t2 r Neq Hr.
  generalize r a1 a2 Neq Hr.
  eapply proj1. eapply proj2. apply substASRQ_fresh_irrel.
Qed.

Corollary substAQ_fresh_irrel : forall a1 t1 a2 t2 q,
  a1 <> a2 ->
  substAQ a1 t1 q = substAQ a2 t2 (substAQ a1 t1 q) ->
  substAQ a2 t2 q = q.
Proof.
  intros a1 t1 a2 t2 q Neq Hq.
  generalize q a1 a2 Neq Hq.
  eapply proj2. eapply proj2. apply substASRQ_fresh_irrel.
Qed.

(*@crusso: not used
Lemma substsAS_fresh_irrel : forall a t as' ts s,
  ~ In a as' ->
  length as' = length ts ->
  substAS a t s = substsAS as' ts (substAS a t s) ->
  substsAS as' ts s = s.
Proof.
  intros a t.
  unfold substsAS.
  induction as'.
    Case "nil".
      intros ts s Nin Len Hs.
      destruct ts; auto.
    Case "a::as'".
      intros ts s Nin Len Hs.
      destruct ts; auto.
      fold substsAS in *.
(*@ continue here *)
Admitted .

Lemma substsAR_fresh_irrel : forall a t as' ts r,
  ~ In a as' ->
  length as' = length ts ->
  substAR a t r = substsAR as' ts (substAR a t r) ->
  substsAR as' ts r = r.
Proof.
Admitted .

Lemma substsAQ_fresh_irrel : forall a t as' ts q,
  ~ In a as' ->
  length as' = length ts ->
  substAQ a t q = substsAQ as' ts (substAQ a t q) ->
  substsAQ as' ts q = q.
Proof.
Admitted .
*)

(*************************************************************************)
(* Opening *)
(*************************************************************************)

Definition openAS t s := substAS 0 t s.

Fixpoint opensAS (ts : list typ) s {struct ts} : sig :=
  match ts with
  | nil => s
  | t::ts' => substAS (length ts') t (opensAS ts' s)
  end.

Definition openAR t s := substAR 0 t s.
Fixpoint opensAR (ts : list typ) r {struct ts} : str :=
  match ts with
  | nil => r
  | t::ts' => substAR (length ts') t (opensAR ts' r)
  end.

Definition openAQ t q := substAQ 0 t q.
Fixpoint opensAQ (ts : list typ) q {struct ts} : qsig :=
  match ts with
  | nil => q
  | t::ts' => substAQ (length ts') t (opensAQ ts' q)
  end.


Lemma opensAS_nil : forall L as' s,
  fresh L (length (nil (A := knd))) as' ->
  opensAS (freeTs as') s = s.
Proof.
  intros L as' s H.
  destruct as'.
    unfold freeTs. simpl map. simpl opensAS. reflexivity.
    simpl fresh in H. inversion H.
Qed.

(*@crusso: not used 
Lemma opensAS_freshes_irrel : forall (a : var) t ts s,
  substAS a t s = opensAS ts (substAS a t s) ->
  opensAS ts s = s.
Proof.
Admitted . 

Lemma opensAR_freshes_irrel : forall (a : var) t ts r,
  substAR a t r = opensAR ts (substAR a t r) ->
  opensAR ts r = r.
Proof.
Admitted .

Lemma opensAQ_freshes_irrel : forall (a : var) t ts q,
  substAQ a t q = opensAQ ts (substAQ a t q) ->
  opensAQ ts q = q.
Proof.
Admitted .
*)


(*************************************************************************)
(* Opening and constructors *)
(*************************************************************************)

Lemma opensAS_val : forall ts t,
  opensAS ts (valS t) = valS (opensAT ts t).
Proof.
  intros ts t.
  induction ts; simpl opensAS; simpl opensAT.
    reflexivity.
    rewrite IHts. simpl substAS. reflexivity.
Qed.

Lemma opensAS_typ : forall ts t k,
  opensAS ts (typS t k) = typS (opensAT ts t) k.
Proof.
  intros ts t k.
  induction ts; simpl opensAS; simpl opensAT.
    reflexivity.
    rewrite IHts. simpl substAS. reflexivity.
Qed.

Lemma opensAS_sig : forall ts q,
  opensAS ts (sigS q) = sigS (opensAQ ts q).
Proof.
  intros ts q.
  induction ts; simpl opensAS; simpl opensAQ.
    reflexivity.
    rewrite IHts. simpl substAS. reflexivity.
Qed.

Lemma opensAS_str : forall ts r,
  opensAS ts (strS r) = strS (opensAR ts r).
Proof.
  intros ts r.
  induction ts; simpl opensAS; simpl opensAR.
    reflexivity.
    rewrite IHts. simpl substAS. reflexivity.
Qed.

Lemma opensAS_fun : forall ts ks s q,
  opensAS ts (funS ks s q) =
  funS ks (substsAS (map (shifts (length ks)) (bnds (length ts))) ts s) (substsAQ (map (shifts (length ks)) (bnds (length ts))) ts q).
Proof.
  intros ts ks s q.
  induction ts; simpl opensAS; simpl opensAQ.
    reflexivity.
    rewrite IHts. simpl substAS. reflexivity.
Qed.

Lemma opensAR_emp : forall ts,
  opensAR ts empR = empR.
Proof.
  intro ts.
  induction ts; simpl opensAR.
    reflexivity.
    rewrite IHts. reflexivity.
Qed.

Lemma opensAR_lab : forall ts l s r,
  opensAR ts (labR l s r) = labR l (opensAS ts s) (opensAR ts r).
Proof.
  intros ts l s r.
  induction ts; simpl opensAR; simpl opensAS.
    reflexivity.
    rewrite IHts. simpl substAR. reflexivity.
Qed.

Lemma opensAR_seq : forall ts r1 r2,
  opensAR ts (seqR r1 r2) = seqR (opensAR ts r1) (opensAR ts r2).
Proof.
  intros ts.
  induction r1; intro r2.
    rewrite opensAR_emp. simpl seqR. reflexivity.
    simpl seqR. rewrite opensAR_lab. rewrite IHr1. rewrite opensAR_lab. simpl seqR. reflexivity.
Qed.

Lemma find_opensAR : forall l ts r,
  find l (opensAR ts r) = 
    match find l r with
    | None => None
    | Some s => Some (opensAS ts s)
    end.
Proof.
  intros l ts.
  induction r.
    rewrite opensAR_emp. simpl. reflexivity.
    rewrite opensAR_lab. simpl. destruct (l === l0).
      reflexivity.
      apply IHr.
Qed.

Lemma opensAR_diff : forall ts r1 r2,
  opensAR ts (diffR r1 r2) = diffR (opensAR ts r1) (opensAR ts r2).
Proof.
  intros ts.
  induction r1; intro r2.
    do 2 rewrite opensAR_emp. simpl diffR. reflexivity.
    rewrite opensAR_lab. simpl diffR. rewrite find_opensAR. destruct (find l r2).
      rewrite IHr1. reflexivity.
      rewrite opensAR_lab. rewrite IHr1. reflexivity.
Qed.


Lemma opensAQ_ex : forall ts ks s,
  opensAQ ts (exQ ks s) = exQ ks (substsAS (map (shifts (length ks)) (bnds (length ts))) ts s).
Proof.
  intros ts ks s.
  induction ts; simpl opensAQ; simpl substsAS.
    reflexivity.
    rewrite IHts. simpl substAQ. reflexivity.
Qed.



Lemma sizeR_opensAR : forall ts r,
  sizeR r = sizeR (opensAR ts r).
Proof.
  intro ts.
  induction r.
    rewrite opensAR_emp. reflexivity.
    rewrite opensAR_lab. simpl sizeR. rewrite IHr. reflexivity.
Qed.


(*@crusso: not used
(*************************************************************************)
(* Local closure *)
(*************************************************************************)

Inductive lclosedS : sig -> Prop :=
  | lclosedS_val : forall t,
      lclosedT t ->
      lclosedS (valS t)
  | lclosedS_typ : forall t k,
      lclosedT t ->
      lclosedS (typS t k)
  | lclosedS_sig : forall q,
      lclosedQ q ->
      lclosedS (sigS q)
  | lclosedS_str : forall r,
      lclosedR r ->
      lclosedS (strS r)
  | lclosedS_fun : forall L ks s q,
      (forall as', fresh L (length ks) as' -> lclosedS (opensAS (freeTs as') s)) ->
      (forall as', fresh L (length ks) as' -> lclosedQ (opensAQ (freeTs as') q)) ->
      lclosedS (funS ks s q)

with lclosedR : str -> Prop :=
  | lclosedR_emp :
      lclosedR empR
  | lclosedR_lab : forall l s r,
      lclosedS s ->
      lclosedR r ->
      lclosedR (labR l s r)

with lclosedQ : qsig -> Prop :=
  | lclosedQ_ex : forall L ks s,
      (forall as', fresh L (length ks) as' -> lclosedS (opensAS (freeTs as') s)) ->
      lclosedQ (exQ ks s)
.

Scheme IIclosedS := Induction for lclosedS Sort Prop
with IIclosedR := Induction for lclosedR Sort Prop
with IIclosedQ := Induction for lclosedQ Sort Prop.
Combined Scheme IIclosedSRQ from IIclosedS, IIclosedR, IIclosedQ.

Lemma substASRQ_lclosed : forall t,
  (forall s, lclosedS s ->
   forall (n : nat), substAS n t s = s) /\
  (forall r, lclosedR r ->
   forall (n : nat), substAR n t r = r) /\
  (forall q, lclosedQ q ->
   forall (n : nat), substAQ n t q = q).
Proof.
  intro t.
  unfold substAS, substAR, substAQ.
  apply IIclosedSRQ.
    Case "valS".
      intros t' Ht n.
      f_equal. apply substAT_lclosed. apply Ht.
    Case "typS".
      intros t' k Ht n.
      f_equal. apply substAT_lclosed. apply Ht.
    Case "sigS".
      intros q Hq IHq n.
      f_equal. apply IHq.
    Case "strS".
      intros r Hr IHr n.
      f_equal. apply IHr.
    Case "funS".
      intros L ks s q Hs IHs Hq IHq n.
      pick_freshes_gen L (length ks) as'.
      f_equal.
(*
      apply (lemma_opensAS_freshes_irrel t 0 as' (S n) s).
        unfold not. intros H'. inversion H'.
        symmetry. apply H0. apply Fr.
*)
Admitted .

Lemma openAS_lclosed : forall t s,
  lclosedS s ->
  openAS t s = s.
Proof.
Admitted .

Lemma opensAS_lclosed : forall s ts,
  lclosedS s ->
  opensAS ts s = s.
Proof.
Admitted .

Lemma lclosedSRQ_trans :
  (forall s,
   lclosedT (transS s) ->
   lclosedS s) /\
  (forall r,
   lclosedTR (transR r) ->
   lclosedR r) /\
  (forall q,
   lclosedT (transQ q) ->
   lclosedQ q).
Proof.
  apply IISRQ.
    Case "val".
      intros t H.
      apply lclosedS_val. apply H.
    Case "typ".
      intros t k H.
      apply lclosedS_typ. eapply lclosedT_typ_inversion. simpl transS in H. apply H.
    Case "sig".
      intros q IHq H.
      apply lclosedS_sig. apply IHq. eapply lclosedT_typ_inversion. simpl transS in H. apply H.
    Case "str".
      intros s IHr H.
      apply lclosedS_str. apply IHr.
(* eapply lclosedT_str_inversion. simpl transS in H. apply H.*)
Admitted .

Corollary lclosedS_trans : forall s,
   lclosedT (transS s) ->
   lclosedS s.
Proof.
  eapply proj1. apply lclosedSRQ_trans.
Qed.

Corollary lclosedR_trans : forall r,
   lclosedTR (transR r) ->
   lclosedR r.
Proof.
  eapply proj1. eapply proj2. apply lclosedSRQ_trans.
Qed.

Corollary lclosedQ_trans : forall q,
   lclosedT (transQ q) ->
   lclosedQ q.
Proof.
  eapply proj2. eapply proj2. apply lclosedSRQ_trans.
Qed.

*)

(*************************************************************************)
(* Commuting translation *)
(*************************************************************************)

Lemma substAT_transSRQ : forall t',
  lclosedT t' ->
  (forall s,
   forall a,
   substAT a t' (transS s) = transS (substAS a t' s)) /\
  (forall r,
   forall a,
   substATR a t' (transR r) = transR (substAR a t' r)) /\
  (forall q,
   forall a,
   substAT a t' (transQ q) = transQ (substAQ a t' q)).
Proof.
  intros t' Ht'.
  apply IISRQ; simpl transS; simpl transR; simpl transQ.
    Case "valS".
      intros t a.
      reflexivity.
    Case "typS".
      intros t k a.
      unfold typT.
      destruct a; simpl.
        reflexivity.
        reflexivity.
    Case "sigS".
      intros q IHq a.
      unfold typT.
      rewrite <- IHq.
      destruct a; simpl.
        reflexivity.
        reflexivity.
    Case "strS".
      intros s IHs a.
      rewrite substAT_str. f_equal.
      apply IHs.
    Case "funS".
      intros l s IHs q IHq a.
      rewrite substAT_alls. f_equal.
      simpl substAT. f_equal.
        apply IHs.
        apply IHq.
    Case "empR".
      intros a.
      unfold substATR. simpl mapR.
      reflexivity.
    Case "labR".
      intros l s IHs r IHr a.
      unfold substATR. simpl mapR. f_equal.
        apply IHs.
        unfold substATR in IHr. apply IHr.
    Case "exQ".
      intros ks s IHs a.
      rewrite substAT_exs. f_equal.
      apply IHs.
Qed.


Corollary substAT_transS : forall a t' s,
  lclosedT t' ->
  substAT a t' (transS s) = transS (substAS a t' s).
Proof.
  intros a t' s Ht'. generalize dependent a. generalize dependent s.
  eapply proj1. apply substAT_transSRQ.
  apply Ht'.
Qed.

Corollary substATR_transR : forall a t' r,
  lclosedT t' ->
  substATR a t' (transR r) = transR (substAR a t' r).
Proof.
  intros a t' r Ht'. generalize dependent a. generalize dependent r.
  eapply proj1. eapply proj2. apply substAT_transSRQ.
  apply Ht'.
Qed.

Corollary substAT_transQ : forall a t' q,
  lclosedT t' ->
  substAT a t' (transQ q) = transQ (substAQ a t' q).
Proof.
  intros a t' q Ht'. generalize dependent a. generalize dependent q.
  eapply proj2. eapply proj2. apply substAT_transSRQ.
  apply Ht'.
Qed.


Corollary openAT_transS : forall t' s,
  lclosedT t' ->
  openAT t' (transS s) = transS (openAS t' s).
Proof.
  intros t' s Ht'. unfold openAT, openAS. apply substAT_transS. apply Ht'.
Qed.

Corollary openATR_transR : forall t' r,
  lclosedT t' ->
  openATR t' (transR r) = transR (openAR t' r).
Proof.
  intros t' r Ht'. unfold openATR, openAR. apply substATR_transR. apply Ht'.
Qed.

Corollary openAT_transQ : forall t' q,
  lclosedT t' ->
  openAT t' (transQ q) = transQ (openAQ t' q).
Proof.
  intros t' q Ht'. unfold openAT, openAQ. apply substAT_transQ. apply Ht'.
Qed.


Lemma substsAT_transS : forall as' ts s,
  lclosedTs ts ->
  substsAT as' ts (transS s) = transS (substsAS as' ts s).
Proof.
  induction as'; intros ts s Hts; simpl substsAT; simpl substsAS.
    reflexivity.
    destruct ts.
      reflexivity.
      rewrite IHas'.
        apply substAT_transS. eapply lclosedTs_head. apply Hts.
        eapply lclosedTs_tail. apply Hts.
Qed.

Lemma substsATR_transR : forall as' ts r,
  lclosedTs ts ->
  substsATR as' ts (transR r) = transR (substsAR as' ts r).
Proof.
  induction as'; intros ts r Hts; simpl substsATR; simpl substsAR.
    reflexivity.
    destruct ts.
      reflexivity.
      rewrite IHas'.
        apply substATR_transR. eapply lclosedTs_head. apply Hts.
        eapply lclosedTs_tail. apply Hts.
Qed.

Lemma substsAT_transQ : forall as' ts q,
  lclosedTs ts ->
  substsAT as' ts (transQ q) = transQ (substsAQ as' ts q).
Proof.
  induction as'; intros ts q Hts; simpl substsAT; simpl substsAQ.
    reflexivity.
    destruct ts.
      reflexivity.
      rewrite IHas'.
        apply substAT_transQ. eapply lclosedTs_head. apply Hts.
        eapply lclosedTs_tail. apply Hts.
Qed.


Lemma opensAT_transS : forall ts s,
  lclosedTs ts ->
  opensAT ts (transS s) = transS (opensAS ts s).
Proof.
  induction ts; intros s Hts.
    reflexivity.
    simpl opensAT. simpl opensAS. rewrite IHts.
      apply substAT_transS. eapply lclosedTs_head. apply Hts.
      eapply lclosedTs_tail. apply Hts.
Qed.

Lemma opensATR_transR : forall ts r,
  lclosedTs ts ->
  opensATR ts (transR r) = transR (opensAR ts r).
Proof.
  induction ts; intros r Hts.
    reflexivity.
    simpl opensATR. simpl opensAR. rewrite IHts.
      apply substATR_transR. eapply lclosedTs_head. apply Hts.
      eapply lclosedTs_tail. apply Hts.
Qed.

Lemma opensAT_transQ : forall ts q,
  lclosedTs ts ->
  opensAT ts (transQ q) = transQ (opensAQ ts q).
Proof.
  induction ts; intros q Hts.
    reflexivity.
    simpl opensAT. simpl opensAQ. rewrite IHts.
      apply substAT_transQ. eapply lclosedTs_head. apply Hts.
      eapply lclosedTs_tail. apply Hts.
Qed.


(*************************************************************************)
(* Fields *)
(*************************************************************************)

Lemma field_subst : forall l s r a t,
  field l s r ->
  field l (substAS a t s) (substAR a t r).
Proof.
  intros l s r a t H.
  induction r; simpl substAR.
    Case "emp".
      inversion H.
    Case "lab".
      unfold field. simpl find. destruct (l === l0); subst.
        unfold field in H. simpl find in H. destruct (l0 === l0).
          inversion H; subst. reflexivity.
          destruct n. reflexivity.
        unfold field in H. simpl find in H. destruct (l === l0).
          destruct (n e).
          apply IHr. apply H.
Qed.

Corollary field_open : forall l s r t,
  field l s r ->
  field l (openAS t s) (openAR t r).
Proof.
  intros l s r t H. unfold openAS, openAR. apply field_subst. apply H.
Qed.

Corollary field_opens : forall l s r ts,
  field l s r ->
  field l (opensAS ts s) (opensAR ts r).
Proof.
  intros l s r ts H.
  induction ts; simpl opensAS; simpl opensAR.
    apply H.
    apply field_subst. apply IHts.
Qed.

Lemma field_trans : forall X s r,
  field X s r ->
  Derived_strT.field X (transS s) (transR r).
Proof.
  intros X s.
  unfold field. unfold Derived_strT.field.
  induction r; intro H; simpl transR; simpl find in *.
    inversion H.
    simpl getR. destruct (X === l).
      inversion H. subst. reflexivity.
      apply IHr. apply H.
Qed.

Lemma ofT_field : forall C r l s,
  ofTR C (transR r) ->
  field l s r ->
  ofT C (transS s) omK.
Proof.
  intros C r l s Hr Hl.
  induction r.
    Case "emp".
      inversion Hl.
    Case "lab".
      inversion Hr; subst.
      inversion Hl. destruct (l === l0).
        inversion H0. subst. apply H2.
        apply IHr.
          inversion Hr; subst. apply H4.
          apply H0.
Qed.
